!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: keyvaluedict.gs,v 1.10 2008-01-09 19:02:34 stever Exp $
!
! Superclass Hierarchy:
!   KeyValueDictionary, AbstractDictionary, Collection, Object.
!
!=========================================================================

removeallmethods KeyValueDictionary
removeallclassmethods KeyValueDictionary

category: 'For Documentation Installation only'
classmethod: KeyValueDictionary
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'KeyValueDictionary is a concrete subclass of AbstractDictionary.  In a
 KeyValueDictionary, keys may be of mixed classes.

 A KeyValueDictionary stores key-value pairs under an index that is generated by
 applying a hash function to the key; it does not use Associations.  The hashing
 improves retrieval speed.  However, you must observe an important restriction:
 after a key/value pair has been added to a KeyValueDictionary, you must not
 modify the key.  Doing so renders the value inaccessible.

 A KeyValueDictionary is also an equality-based collection.  That is, two keys
 or two values are considered to be the same if they are equivalent; they need
 not be identical to be the same.  Thus, if you add two key-value pairs to a
 KeyValueDictionary but the keys are equivalent, even if they are not identical,
 then the result is that the second pair overwrites the first one, because the
 keys are the same.

 Some other kinds of dictionaries are identity-based rather than equality-based.
 These other kinds of dictionaries exhibit better performance than
 KeyValueDictionary and are to be preferred where they are appropriate.

 For multiuser applications that involve a lot of concurrent use of
 dictionaries, use RcKeyValueDictionary.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A SmallInteger that represents the number of key/value pairs in the
 dictionary.'.
doc documentInstVar: #numElements with: txt.

txt := (GsDocText new) details:
'A SmallInteger that represents the cumulative number of collisions that have
 occurred during the addition of the elements to the dictionary since the last
 rebuild.'.
doc documentInstVar: #numCollisions with: txt.

txt := (GsDocText new) details:
'A SmallInteger that represents the number of collisions allowed before
 rebuilding the hash table.'.
doc documentInstVar: #collisionLimit with: txt.

txt := (GsDocText new) details:
'A SmallInteger that represents the size of the primary hash table.'.
doc documentInstVar: #tableSize with: txt.

self description: doc.
%

category: 'Private'
method: KeyValueDictionary
rebuildIfNeeded

"Private.  If the receiver has too many collisions, rebuild it with a larger
 hash table.  Returns true if the receiver was rebuilt, false otherwise."

| newSize |

numCollisions > collisionLimit ifTrue:[
  [ numCollisions > collisionLimit ] whileTrue: [
    newSize := 
       Integer _selectedPrimeGreaterThan: ( numElements max: tableSize + 1) .
    self rebuildTable: newSize .
    ].
  ^ true .
  ].
^ false
%

category: 'Private'
method: KeyValueDictionary
_rebuild

"Rebuild to the next higher size because the internal data structures like
 the collision buckets would like it to happen."

  | newSize |

  newSize := Integer 
             _selectedPrimeGreaterThan: ( numElements max: tableSize + 1) .
  self rebuildTable: newSize .
  ^ self.
%

category: 'Instance Creation'
classmethod: KeyValueDictionary
new

"Creates an instance of KeyValueDictionary with a default table size."

^self new: 5
%

category: 'Instance Creation'
classmethod: KeyValueDictionary
new: tableSize

"Creates an instance of KeyValueDictionary based upon the specified table size."

| newSize newDict |

newSize := Integer _selectedPrimeGreaterThan: tableSize.
newDict := super _basicNew: (newSize * 2).

newDict _initializeWithoutClear: newSize.
^ newDict.
%

category: 'Enumerating'
method: KeyValueDictionary
keysAndValuesDo: aBlock

"Evaluates aBlock with each of the receiver's key/value pairs as the
 arguments.  The argument aBlock must be a two-argument block.  The
 first argument is the key and the second argument is the value of
 each key/value pair."

| aKey collisionBkt |

1 to: tableSize * 2 by: 2 do: [ :tableIndex |
  nil == (aKey := self _at: tableIndex) ifTrue: [
    nil == (collisionBkt := self _at: (tableIndex + 1)) ifFalse: [
      1 to: collisionBkt _basicSize by: 2 do: [ :j |
        nil == (aKey := collisionBkt _at: j) ifFalse: [
          aBlock value: aKey value: (collisionBkt _at: j + 1).
          ].
        ].
      ].
    ]
  ifFalse: [
    aBlock value: aKey value: (self _at: tableIndex + 1).
    ].
  ].
%

category: 'Enumerating'
method: KeyValueDictionary
do: aBlock

"(R) Iteratively evaluates the one argument block aBlock, using the value of
 each Association or key-value pair as the argument of the block."

self keysAndValuesDo: [:aKey :aValue | aBlock value: aValue].
^self
%

category: 'Enumerating'
method: KeyValueDictionary
associationsDo: aBlock

"Evaluates aBlock with each of the receiver's key/value pairs as the
 argument by creating an Association for each key/value pair.  The
 argument aBlock must be a one-argument block."

self keysAndValuesDo: [:aKey :aValue |
   aBlock value: (Association newWithKey: aKey value: aValue)
].
^self
%

category: 'Enumerating'
method: KeyValueDictionary
keysDo: aBlock

"For each key/value pair in the receiver, evaluates the one-argument block
 aBlock with the key as the argument."

self keysAndValuesDo: [:aKey :aValue |
   aBlock value: aKey
].
^self
%

category: 'Enumerating'
method: KeyValueDictionary
valuesDo: aBlock

"For each key/value pair in the receiver, evaluates the one-argument block
 aBlock with the value as the argument."

self keysAndValuesDo: [:aKey :aValue |
   aBlock value: aValue
].
^self
%

! _asCollection removed in v4.0

category: 'Accessing'
method: KeyValueDictionary
keyAtValue: anObject ifAbsent: aBlock

"(R) Returns the key of the first value equal to the given object, anObject.
 If no match is found, evaluates and returns the result of the block aBlock."

"Note: In GemStone 4.1, this method checked for a value identical to anObject
 as opposed to checking for equality."

| aKey collisionBkt |

1 to: tableSize * 2 by: 2 do: [ :tableIndex |
  nil == (aKey := self _at: tableIndex) ifTrue: [
    nil == (collisionBkt := self _at: (tableIndex + 1)) ifFalse: [
      1 to: collisionBkt _basicSize by: 2 do: [ :j |
        nil == (aKey := collisionBkt _at: j) ifFalse: [
          anObject = (collisionBkt _at: j + 1) ifTrue: [
            ^aKey
            ].
          ].
        ].
      ].
    ]
  ifFalse: [
    anObject = (self _at: tableIndex + 1) ifTrue: [
      ^aKey
      ].
    ].
  ].

^aBlock value.
%

category: 'Removing'
method: KeyValueDictionary
removeKey: aKey ifAbsent: aBlock

"Removes the key/value pair with key aKey from the receiver and returns
 the value.  If no key/value pair is present with key aKey, evaluates
 the zero-argument block aBlock and returns the result of that evaluation."

| hash thisKey collisionBkt aValue lastPair |

hash := self hashFunction: aKey.
nil == (thisKey := (self keyAt: hash))

   ifTrue: [ "It might be in a collision bucket"
      collisionBkt := self valueAt: hash.
      nil == collisionBkt
         ifTrue: [ ^ self _reportKeyNotFound: aKey with: aBlock ] .
      aValue := collisionBkt
         removeKey: aKey
         ifAbsent: [ ^ self _reportKeyNotFound: aKey with: aBlock]. "fix 11884"
      "If we got this far, a key/value pair was removed from collisionBkt"
      numCollisions := numCollisions - 1.
      (collisionBkt size <= 1) ifTrue: [
         "We don't need the collision bucket anymore.
          Put its last key/value pair in our table."
         lastPair := collisionBkt firstPair.
         self atHash: hash putKey: lastPair key.
         self atHash: hash putValue: lastPair value.
         collisionBkt _removeAll ]]

   ifFalse: [ "There is a key/value pair in this slot"
      (self compareKey: aKey with: thisKey)
         ifFalse: [ ^ self _reportKeyNotFound: aKey with: aBlock ] .
      "The key matches - remove it"
      aValue := self valueAt: hash.
      self atHash: hash putKey: nil.
      self atHash: hash putValue: nil ].

numElements := numElements - 1.
^aValue
%

category: 'Private'
method: KeyValueDictionary
_removeKey: aKey ifAbsent: aBlock

"Removes the key/value pair with key aKey from the receiver and returns
 the value.  If no key/value pair is present with key aKey, evaluates
 the zero-argument block aBlock and returns the result of that evaluation."

"Reimplemented for direct inheritance by IdentityDictionary and below.
 Do not reimplement below KeyValueDictionary."

| hash thisKey collisionBkt aValue lastPair |

hash := self hashFunction: aKey.
nil == (thisKey := (self keyAt: hash))

   ifTrue: [ "It might be in a collision bucket"
      collisionBkt := self valueAt: hash.
      nil == collisionBkt
         ifTrue: [ ^ self _reportKeyNotFound: aKey with: aBlock ] .
      aValue := collisionBkt
         removeKey: aKey
         ifAbsent: [ ^ self _reportKeyNotFound: aKey with: aBlock]. "fix 11884"
      "If we got this far, a key/value pair was removed from collisionBkt"
      numCollisions := numCollisions - 1.
      (collisionBkt size <= 1) ifTrue: [
         "We don't need the collision bucket anymore.
          Put its last key/value pair in our table."
         lastPair := collisionBkt firstPair.
         self atHash: hash putKey: lastPair key.
         self atHash: hash putValue: lastPair value ]]

   ifFalse: [ "There is a key/value pair in this slot"
      (self compareKey: aKey with: thisKey)
         ifFalse: [ ^ self _reportKeyNotFound: aKey with: aBlock ] .
      "The key matches - remove it"
      aValue := self valueAt: hash.
      self atHash: hash putKey: nil.
      self atHash: hash putValue: nil ].

numElements := numElements - 1.
^aValue
%

category: 'Accessing'
method: KeyValueDictionary
collisionLimit

"Returns the value of the collisionLimit instance variable."

^collisionLimit
%

category: 'Private'
method: KeyValueDictionary
keyAt: hashOffset

"Private.  Returns the key at the specified hash offset."

^ self _at: (2 * hashOffset - 1)
%

category: 'Accessing'
method: KeyValueDictionary
tableSize

"Returns the size of the primary hash table."

^tableSize
%

category: 'Accessing'
method: KeyValueDictionary
numCollisions

"Returns the total number of collisions, the sum of the number of key/value
 pairs in all collision buckets."

^ numCollisions
%

category: 'Private'
method: KeyValueDictionary
valueAt: hash

"Returns the value at the specified hash offset."

^self _basicAt: (2 * hash)
%

category: 'Accessing'
method: KeyValueDictionary
size

"(R) Returns the number of key/value pairs in the receiver."

^numElements
%

category: 'Accessing'
method: KeyValueDictionary
numElements

"Same as the size method."

^numElements
%

category: 'Accessing'
method: KeyValueDictionary
at: aKey ifAbsent: aBlock

"Returns the value that corresponds to aKey.  If no such key/value pair
 exists, returns the result of evaluating the zero-argument block aBlock."

| hash hashKey collisionBkt |

nil == aKey ifTrue:[ ^self _reportKeyNotFound: aKey with: aBlock ].
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifFalse: [ ^collisionBkt at: aKey ifAbsent: aBlock ]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ ^self valueAt: hash ] ].
^ self _reportKeyNotFound: aKey with: aBlock
%

category: 'Accessing'
method: KeyValueDictionary
keys

"Returns an IdentitySet containing the receiver's keys."

| result |
result := IdentitySet new .
self keysDo:[ :aKey | result add: aKey ].
^ result
%

category: 'Private'
method: KeyValueDictionary
atHash: hashIndex putKey: aKey

"Updates the hash table by storing aKey at the specified hashIndex.

 Note: This method overwrites the key value at the given hashIndex and
 destroys the ordering of the keys in the receiver.  Returns aKey."

((hashIndex <= 0) | (hashIndex > tableSize))
  ifTrue: [hashIndex _error: #rtErrArgOutOfRange ].
self _basicAt: (2 * hashIndex - 1) put: aKey
%

category: 'Private'
method: KeyValueDictionary
atHash: hashIndex putValue: aValue

"Updates the hash table by storing aValue at the specified hashIndex.
 Returns aValue."

((hashIndex <= 0) | (hashIndex > tableSize))
  ifTrue: [ ^ hashIndex _error: #rtErrArgOutOfRange ].
^self _basicAt: (2 * hashIndex) put: aValue
%

category: 'Updating'
method: KeyValueDictionary
collisionLimit: aLimit

"Sets the collision limit to aLimit."

"the value 536870911 is special and may only be installed by instance methods.
 This value is used to prevent rehash during symbol canonicalization."

aLimit == 536870911 
  ifTrue:[ collisionLimit := aLimit - 1]
  ifFalse:[ collisionLimit := aLimit ]
%

category: 'Private'
method: KeyValueDictionary
_at: aKey put: aValue

"Stores the aKey/aValue pair in the hash dictionary.  Rebuilds the hash table
 if the addition caused the number of collisions to exceed the limit allowed."

"Duplicate of at:put: for direct inheritance by subclasses of
 IdentityDictionary.  Do not reimplement in subclasses of KeyValueDictionary."

| hash hashKey hashValue collisionBkt saveBktSize|
nil == aKey ifTrue: [ ^ self _error: #rtErrNilKey ].
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifTrue: [ self atHash: hash putKey: aKey.
                        self atHash: hash putValue: aValue.
                        numElements := numElements + 1 ]
              ifFalse: [ saveBktSize := collisionBkt size.
                         collisionBkt at: aKey put: aValue keyValDict: self.
                         (saveBktSize == collisionBkt size)
                           ifFalse: [
                             numElements := numElements + 1.
                             numCollisions := numCollisions + 1
                           ].
                         numElements := numElements]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ self atHash: hash putKey: aKey.
                         self atHash: hash putValue: aValue ]
               ifFalse: [ numElements := numElements + 1.
                          hashValue := self valueAt: hash.
                          collisionBkt := self collisionBucketClass new.
                          collisionBkt assignToSegment: self segment.
                          collisionBkt keyValueDictionary: self.
                          self atHash: hash putKey: nil.
                          self atHash: hash putValue: collisionBkt.
                          collisionBkt at: hashKey put: hashValue keyValDict: self.
                          collisionBkt at: aKey put: aValue keyValDict: self.
                          numCollisions := numCollisions + 1 ]
           ].
(numCollisions > collisionLimit)
  ifTrue: [
    self rebuildTable: (Integer _selectedPrimeGreaterThan: tableSize * 2)
  ].
^aValue
%

category: 'Updating'
method: KeyValueDictionary
at: aKey put: aValue

"Stores the aKey/aValue pair in the hash dictionary.  Rebuilds the hash table
 if the addition caused the number of collisions to exceed the limit allowed."

"Code is duplicated in _at:put:"

| hash hashKey hashValue collisionBkt saveBktSize|
nil == aKey ifTrue: [ ^ self _error: #rtErrNilKey ].
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifTrue: [ self atHash: hash putKey: aKey.
                        self atHash: hash putValue: aValue.
                        numElements := numElements + 1 ]
              ifFalse: [ saveBktSize := collisionBkt size.
                         collisionBkt at: aKey put: aValue keyValDict: self.
                         (saveBktSize == collisionBkt size)
                           ifFalse: [
                             numElements := numElements + 1.
                             numCollisions := numCollisions + 1
                           ].
                         numElements := numElements]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ self atHash: hash putKey: aKey.
                         self atHash: hash putValue: aValue ]
               ifFalse: [ numElements := numElements + 1.
                          hashValue := self valueAt: hash.
                          collisionBkt := self collisionBucketClass new.
                          collisionBkt assignToSegment: self segment.
                          collisionBkt keyValueDictionary: self.
                          self atHash: hash putKey: nil.
                          self atHash: hash putValue: collisionBkt.
                          collisionBkt at: hashKey put: hashValue keyValDict: self.
                          collisionBkt at: aKey put: aValue keyValDict: self.
                          numCollisions := numCollisions + 1 ]
           ].
(numCollisions > collisionLimit)
  ifTrue: [
    self rebuildTable: (Integer _selectedPrimeGreaterThan: tableSize * 2)
  ].
^aValue
%

category: 'Clustering'
method: KeyValueDictionary
clusterDepthFirst

"This method clusters the receiver and its named instance variables and the
 key/value pairs in depth-first order.  Returns true if the receiver has
 already been clustered during the current transaction; returns false
 otherwise."

self cluster ifTrue: [ ^ true ].

1 to: self class instSize do:[ :i | 
  (self instVarAt: i) clusterDepthFirst 
  ].
1 to: tableSize * 2 by: 2 do: [ :tableIndex | | aKey collisionBkt |
  nil == (aKey := self _at: tableIndex) ifTrue: [
    nil == (collisionBkt := self _at: (tableIndex + 1)) ifFalse: [
      collisionBkt cluster .
      1 to: collisionBkt _basicSize by: 2 do: [ :j |
        nil == (aKey := collisionBkt _at: j) ifFalse: [
          aKey clusterDepthFirst. 
          (collisionBkt _at: j + 1) clusterDepthFirst.
          ].
        ].
      ].
    ]
  ifFalse: [
    aKey clusterDepthFirst .
    (self _at: tableIndex + 1) clusterDepthFirst.
    ].
  ].
^ false
%

category: 'Copying'
method: KeyValueDictionary
copy

"Returns a copy of the receiver."

|newKeyValueDict collisionBkt|
newKeyValueDict := super copy.
1 to: tableSize do:
  [ :index | nil == (self keyAt: index)
      ifTrue: [ collisionBkt := self valueAt: index.
         nil == collisionBkt
           ifFalse: [ collisionBkt := (self valueAt: index) copy.
                      collisionBkt keyValueDictionary: newKeyValueDict.
                      newKeyValueDict atHash: index putValue: collisionBkt. ]  ]
  ].
^newKeyValueDict
%

! selectValuesAsArray: inherited
! select:  inherited
! collect: inherited
! inherited: detectValues: aBlock ifNone: exceptionBlock

category: 'Hashing'
method: KeyValueDictionary
rebuildTable: newSize

"Rebuilds the hash table by saving the current state, initializing and
 changing the size of the table, and adding the key value pairs saved
 back to the hash dictionary."

|saveTable index saveCollLimit|

collisionLimit == 536870911 ifTrue:[
  "rehash disabled during symbol canonicalization."
  ^ self 
  ].

saveTable := Array new: (self size * 2).
index := 0.
self keysAndValuesDo: [ :aKey :aValue |  
  index := index + 1.
  saveTable at: index put: aKey.
  index := index + 1.
  saveTable at: index put: aValue  
  ].
self tableSize: newSize.
saveCollLimit := collisionLimit .
collisionLimit := 536870911 . "prevent recursive rebuild"
1 to: index by: 2 do: [ :i |
  self at: (saveTable at: i "key") put: (saveTable at: i + 1 "value") 
  ].
collisionLimit := saveCollLimit .
saveTable _basicSize: 0 . "reduce garbage"
^self
%

category: 'Initializing'
method: KeyValueDictionary
_gciInitialize

"Private."

tableSize == nil
  ifTrue: [ self initialize: 5 ]
  ifFalse: [ self initialize: tableSize ]
%

category: 'Private'
method: KeyValueDictionary
_initializeWithoutClear: newSize

"Private. Initializes the instance variables of the receiver to be an empty
 KeyValueDictionary of the specified size. Does not clear the contents
 of the receiver - assumes they are all nil."

tableSize := newSize.
numElements := 0.
numCollisions := 0.
collisionLimit := newSize // 2.

^self
%

category: 'Initializing'
method: KeyValueDictionary
initialize: itsSize

"Initializes the instance variables of the receiver to be an empty
 KeyValueDictionary of the specified size."

| newSize |

newSize := itsSize .

(newSize <= 0)
  ifTrue: [
    newSize _error: #rtErrArgNotPositive.  
    newSize := 3 
    ].

self _basicSize: 0.  "Set size to 0 and restore to nil any entries"
self _basicSize: newSize * 2. 
self _initializeWithoutClear: newSize.
^self
%

category: 'Initializing'
method: KeyValueDictionary
tableSize: newSize

"Sets the table size to a new value."

"When used during rehashing a dictionary the sender must have already
 copied all key/value pairs to a temporary object to save the contents
 of the dictionary."

| physSize aBucket |

(newSize <= 0)
  ifTrue: [newSize _error: #rtErrArgNotPositive  .  ^ self ].

"dereference old collision buckets from their parent to avoid problems
 when resolving symbol canonicalization at commit."
1 to: self _basicSize by: 2 do:[:j |
  (self basicAt: j) == nil ifTrue:[ "no key present, so check for a bucket"
    (aBucket := self basicAt: j + 1) ~~ nil ifTrue:[
       aBucket _removeAll . "dereference from parent"
      ].
    ].
  ].

physSize := newSize * 2 .
self nilOutAll .
self _basicSize: physSize .
numElements := 0.
numCollisions := 0.
collisionLimit := newSize // 4.
tableSize := newSize .
%

category: 'Statistics'
method: KeyValueDictionary
statistics

"Returns a Dictionary containing statistics that can be useful in determining
 the performance of a hash dictionary.  The dictionary contains the following
 information:

 TotalCollisionBuckets - The number of collision buckets required to
                         implement the hash dictionary.

 AveragePairsPerBucket - The average number of key/value pairs in each bucket.

 LargestBucket         - The bucket having the most key/value pairs.  This
                         bucket contains the most keys for which the hash
                         function did not provide a good distribution over the
                         range of values in the table."

| hashKey collisionBkt bktCount pairCount maxBkt maxBktSize result|

bktCount := 0.
pairCount :=0.
maxBktSize := 0.
1 to: tableSize do: [ :tableIndex |
  hashKey := self keyAt: tableIndex.
  collisionBkt := self valueAt: tableIndex.
  (nil == hashKey _and:[ nil ~~ collisionBkt])
    ifTrue: [ bktCount := bktCount + 1.
              pairCount := pairCount + collisionBkt size.
              (collisionBkt size > maxBktSize)
                 ifTrue: [ maxBktSize := collisionBkt size.
                           maxBkt := collisionBkt] ]
  ].
result := SymbolDictionary new.
result at: #TotalCollisionBuckets put: bktCount.
(bktCount > 0)
  ifTrue: [result at: #AveragePairsPerBucket put: pairCount // bktCount]
  ifFalse: [result at: #AveragePairsPerBucket put: 0].
result at: #LargestBucket put: maxBkt.
^result
%

category: 'Private'
method: KeyValueDictionary
_at: aKey otherwise: defaultValue

"Returns the value that corresponds to aKey.  If no such key/value pair
 exists, returns the given alternate value."

"Duplicate of at:otherwise: for direct inheritance by things 
 below IdentityDictionary.  Do not reimplement below KeyValueDictionary."

| hash hashKey collisionBkt |

nil == aKey ifTrue:[ ^ defaultValue ].
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifFalse: [ ^collisionBkt at: aKey otherwise: defaultValue ]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ ^self valueAt: hash ] ].
^ defaultValue
%

category: 'Accessing'
method: KeyValueDictionary
at: aKey otherwise: aValue

"Returns the value that corresponds to aKey.  If no such key/value pair
 exists, returns the given alternate value."

| hash hashKey collisionBkt |

nil == aKey ifTrue:[ ^ aValue ].
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifFalse: [ ^collisionBkt at: aKey otherwise: aValue ]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ ^self valueAt: hash ] ].
^ aValue
%

category: 'Accessing'
method: KeyValueDictionary
keyAt: aKey otherwise: aValue

"Returns the key that corresponds to aKey.  If no such key/value pair
 exists, returns the given alternate value.

 Note that the keyAt: method is private and does not have behavior that is
 compatible with this method."

| hash hashKey collisionBkt |

nil == aKey ifTrue:[ ^ self _error: #rtErrNilKey ] .
hash := self hashFunction: aKey.
hashKey := self keyAt: hash.
nil == hashKey
  ifTrue: [ collisionBkt := self valueAt: hash.
            nil == collisionBkt
              ifFalse: [ ^ collisionBkt keyAt: aKey otherwise: aValue ]
          ]
  ifFalse: [ (self compareKey: aKey with: hashKey)
               ifTrue: [ ^ hashKey ] ].
^ aValue
%

category: 'Private'
method: KeyValueDictionary
collisionBucketClass

"Returns the class of object to create when keys collide."

^ CollisionBucket
%

category: 'Private'
method: KeyValueDictionary
compareKey: key1 with: key2

"Returns whether key1 is equivalent to key2."

^ key1 = key2
%

category: 'Locking Support'
method: KeyValueDictionary
_lockableValues

"Returns an Array of the receiver's keys and values."

| result |

result := Array new .
self keysAndValuesDo: [:aKey :aValue |
  result add: aKey .
  result add: aValue .
] .
^ result
%

category: 'Searching'
method: KeyValueDictionary
_idxOccurrencesOf: aValue

"Dictionaries can only contain Associations."

^ 0
%

category: 'Copying'
method: KeyValueDictionary
_basicCopyFrom: index1 to: index2 into: destObj startingAt: destIndex

"Copies the elements of the receiver between index1 and index2, inclusive, into
 destObj starting at destIndex, overwriting the previous contents.  If
 destObj is the same object as the receiver, the source and destination blocks
 may overlap.  destObj may be either an Array or a KeyValueDictionary  "

<primitive: 807"with rcvrClass">

(index1 > index2)
  ifTrue:[ index1 _error: #rtErrBadCopyFromTo args: #[index2]].

(index1 < 1) ifTrue:[ self _errorIndexOutOfRange: index1].

(index2 > self size) ifTrue:[ self _errorIndexOutOfRange: index2].

((destIndex < 1) _or: [(destIndex > (destObj size + 1))])
  ifTrue:[ self _errorIndexOutOfRange: destIndex].
destObj _validateClasses: #[ Array, KeyValueDictionary ] .
self _primitiveFailed: #_basicCopyFrom:to:into:startingAt: 
%

! remove duplicate implementation of firstPublic...

category: 'Updating'
method: KeyValueDictionary
changeToSegment: segment

"(R) Assigns the receiver and all its components to the given segment. Returns
 the receiver."

| collisionBkt |

self assignToSegment: segment.
1 to: tableSize * 2 by: 2 do: [ :tableIndex |
  ( nil == (self _at: tableIndex) _and:
  [ (collisionBkt := self _at: (tableIndex + 1)) ~~ nil ] )
    ifTrue: [ collisionBkt assignToSegment: segment ]
].
^ self.
%

category: 'Accessing the Class Format'
classmethod: KeyValueDictionary
firstPublicInstVar

"Returns the index of the first user-visible instance variable defined in this
 class, whether or not this class actually has user-visible instance variables."

^ 5
%

category: 'Storing and Loading'
classmethod: KeyValueDictionary
loadFrom: passiveObj mappingToClass: newClass

"Reads from passiveObj the passive form of an object with named instance
 variable format.  Converts the object to its active form by loading the
 information into a new instance of the receiver.  Returns the new instance."

| newDict oldDict size |
(passiveObj version >= 500 _or:[ newClass == self]) ifTrue:[ 
  size := passiveObj readSize.
  newDict := self new.
  newDict loadFrom: passiveObj size: size.
  ]
ifFalse:[
  "handle activation of objects written by 4.1.3"

  (self instSize - self firstPublicInstVar) > 
  (newClass instSize - newClass firstPublicInstVar) ifTrue:[
    self _halt:'Unable to map subclass of ' + self name +
               ' to class ' + newClass name  .
    ].
  newDict := newClass new .
  passiveObj hasRead: newDict .
  size := passiveObj readSize. 
  oldDict := self new .
  oldDict basicLoadFromNoRead: passiveObj size: size .
  oldDict keysAndValuesDo:[:aKey :aValue | newDict at: aKey put: aValue ].
  oldDict initialize: 1 .
].
^ newDict
%

category: 'Storing and Loading'
method: KeyValueDictionary
loadVaryingFrom: passiveObj size: varyingSize

"Reads the varying part of the receiver from the given passive object.
 Does not record the receiver as having been read.  Does not read the
 receiver's named instance variables, if any."

passiveObj version >= 500 ifTrue:[
  1 to: varyingSize do: [:i |
    self at: passiveObj readObject"aKey" put: passiveObj readObject"aValue"
    ].
  ] 
ifFalse:[ super loadVaryingFrom: passiveObj size: varyingSize ].
%

category: 'Storing and Loading'
method: KeyValueDictionary
basicWriteTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

| s cls ivs c |
  cls := self class.
  passiveObj writeClass: cls.

  ivs := cls _instVarNames .

  passiveObj writeSize: (s := self size) .

  cls firstPublicInstVar to: cls instSize do: [:i |
    (self shouldWriteInstVar: (ivs at: i)) ifTrue: [
      passiveObj writeObject: (self instVarAt: i) named: (ivs at: i)
    ].
  ].

  passiveObj endNamedInstVars.

  c := 0.
  self keysAndValuesDo:[ :aKey :aValue | 
    passiveObj writeObject: aKey; writeObject: aValue .
    c := c + 1.
    c > 49 ifTrue: [
      passiveObj lf.
      c := 0.
      ].
    ].

  passiveObj cr
%

category: 'Private'
method: KeyValueDictionary
_deferredGciUpdateWith: valueArray

"Private."

1 to: valueArray size by: 2 do:[:j |
  self at: (valueArray at: j) put: (valueArray at: j + 1)
  ].
%

category: 'Private'
method: KeyValueDictionary
_collisionBucketsDo: aBlock

"Private.  Executes the given one-argument block for each collision bucket
 in the receiver."

tableSize == nil ifFalse: [
  1 to: tableSize * 2 by: 2 do: [ :tableIndex |
    nil == (self _at: tableIndex)
      ifTrue: [ | collisionBkt |
        nil == (collisionBkt := self _at: (tableIndex + 1))
          ifFalse: [ aBlock value: collisionBkt ]
      ]
    ]
]
%

category: 'Storing and Loading'
method: KeyValueDictionary
loadNamedIVsFrom: passiveObj

"Reads named instance variables from the given passive object.  The
 first instance variable should already have been parsed and be
 available in the passiveObj argument."

| name offset nameSym aValue |

passiveObj version >= 500 ifTrue:[ ^ super loadNamedIVsFrom: passiveObj ].

[ name := passiveObj ivName.
  name ~~ nil ifTrue: [
    nameSym := Symbol _existingWithAll: name .
    nameSym ~~ nil ifTrue:[
      offset := self class _ivOffsetOf: nameSym.
      offset ~~ nil ifTrue:[
        aValue := passiveObj ivValue .
        offset == 3 "collisionLimit" ifTrue:[
          aValue := (tableSize * 100) min: 500000000 .
          ].
        self instVarAt: offset put: aValue .
        ]
      ifFalse:[ self obsoleteInstVar: nameSym value: passiveObj ivValue].
    ]
    ifFalse:[ 
      self obsoleteInstVar: name value: passiveObj ivValue
    ].
    passiveObj readNamedIV
  ]
  ifFalse: [
    false
  ]
] untilFalse.

passiveObj skipNamedInstVars.
%

category: 'Repository Conversion'
method: KeyValueDictionary
rehashForConversion

"Private. Rehashes the receiver because the hash values of some of its keys
 may have changed.  Returns true if the rehashing is successful.  Returns false
 otherwise."

| newSize |

newSize := (Integer _selectedPrimeGreaterThan: numElements) max: 
           (Integer _selectedPrimeGreaterThan: (tableSize - 1)).
"The following helps bypass a bug in one customer's repository."
(numElements < 0) ifTrue: [ numElements := 0 ].
self rebuildTable: newSize .
^ true.
%

category: 'Searching'
method: KeyValueDictionary
includesIdenticalAssociation: anAssociation

"(R) Returns true if anAssociation is identical to one of the Associations of 
 the receiver.  Returns false otherwise.

 KeyValueDictionaries do not store Associations. So this message is not
 meaningful for this class."

^ self shouldNotImplement: #includesIdenticalAssociation.
%

category: 'Private'
method: KeyValueDictionary
_resetParentRef

"Private. After a become:, the parent refs of the collisionBuckets must 
 be reset to point to the correct parent."

self _collisionBucketsDo: [:collisionBkt |
  collisionBkt keyValueDictionary: self].
%

category: 'Private'
method: KeyValueDictionary
auditDictionary

"Audit the structure of the receiver.  Verifies the numElements instance
 variable of the receiver and all collision buckets is correct.  Does not
 verify the reachability of keys.

 Returns an Array with 2 elements. The first element will be true if the audit
 passed or a string describing the size discrepency if the audit failed.  The
 second element will be an IdentitySet which contains any collision buckets which
 did not pass the audit."

| totalSize badBuckets result |
totalSize := 0.
badBuckets := IdentitySet new.
1 to: tableSize * 2 by: 2 do: [ :tableIndex | | key |
  key := self _at: tableIndex .
  key == nil "collision bucket or no entry"
   ifTrue:[ | bucket |
     bucket := self _at: (tableIndex + 1) .
     bucket == nil
       ifFalse:[ totalSize := totalSize + (bucket auditFor: self with: badBuckets) ].
   ] ifFalse:[ totalSize := totalSize + 1 ].
].

result := Array new: 2.
totalSize = numElements
  ifTrue:[ result at: 1 put: true ]
  ifFalse:[ |str|
    str := String new.
    str addAll: 'Error: numElements is '; 
        addAll: numElements asString ;
        addAll: ' but audit found ';
        addAll: totalSize asString;
        addAll: ' elements.'.
     result at: 1 put: str.
  ].
  result at: 2 put: badBuckets.
^result
%
